{
    
    Copyright (c) 2010 Ben Golightly

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    ---------------------------------------------------------------------------
    
    Built in two days for the SoCoder.net Xmas Xomo 2010
    
    Requires lexini (open source) for reading cfg.ini
    http://www.tophatstuff.co.uk/?p=95
    
    Compile with Cobra
    http://www.cobrabytes.com/
    
}
Program (icon:"rsc\jigsaw_ico.ico")
    Uses
        cobra2d,
        keyset,
        lexini

Const
    TI = 12
    PP = 10
    MAX_PARTS = 64
    PZL_PADDING = 128

Type pieces = record
    e: Element
    ew, eh: Integer
    targetx, targety: Integer
    posx, posy: Integer
    solved: Boolean
    
    outie_r: Boolean
    outie_d: Boolean
EndType

Type parts = record
    e: Element
    xvol, yvol, xpos, ypos: Real
    size, ealpha, rot: Real
    active: Boolean
EndType

Type puzzles = record
    e, e_hover, e_reflection: Element
    filename: String
    w, h: Integer
    x, scale, ealpha: Real
EndType

Var
    SW: Integer = 1024
    SH: Integer = 768

    gFullscreen: Boolean = FALSE
    gScreenDepth: Integer = 32
    gShowSplash: Boolean = TRUE
    gParticles: Boolean = TRUE
    
    template: Array [TI] of Element
    template_scaled: Array [TI] of Element
    template_flipped: Array [TI] of Element
    previewPiece: Array [PP] of Element
    
    canvas: Element
    particleMaster: Element
    corner: Element
    btn: Array[3,3] of Element
    
    piece: ^pieces
    pieceList: List of pieces
    part: ^parts
    partList: List of parts
    puzzle: ^puzzles
    puzzleList: List of puzzles

Procedure InitCfg()
Var
    fin: ^iniFiles
Begin
    If (not FileExists("cfg.ini")) then exit
    fin = IniLoad("cfg.ini")

    If (not IniSelectSection(fin, "options")) then exit
    
    SW = IniAsInteger("screen_width")
    SH = IniAsInteger("screen_height")
    gFullscreen = (IniAsInteger("fullscreen") = 1)
    gScreenDepth = IniAsInteger("screen_depth")
    gShowSplash = (IniAsInteger("show_splash_intro") = 1)
    gParticles = (IniAsInteger("particles") = 1)
    
    IniFree(fin)
End

Procedure InitScreen()
Begin
    If (SW = 0) then SW = SelWidth()
    If (SH = 0) then SH = SelHeight()

    OpenScreen(SW, SH, gScreenDepth, gFullscreen, COB_SHOWBORDER)
    VSync(1) 
    canvas = CreateSprite(SW, SH)
    CLS(ToRGBA(0,0,0), canvas)
End

Procedure InitTemplate()
Begin
    template[0] = LoadImage("rsc/template1.png")
    template[1] = NewRotatedTemplateImage(template[0],  90)
    template[2] = NewRotatedTemplateImage(template[0], 180)
    template[3] = NewRotatedTemplateImage(template[0], 270)
    template[4] = LoadImage("rsc/template2.png")
    template[5] = NewRotatedTemplateImage(template[4],  90)
    template[6] = NewRotatedTemplateImage(template[4], 180)
    template[7] = NewRotatedTemplateImage(template[4], 270)
    template[8] = LoadImage("rsc/template3.png")
    template[9] = NewRotatedTemplateImage(template[8],  90)
    template[10]= NewRotatedTemplateImage(template[8], 180)
    template[11]= NewRotatedTemplateImage(template[8], 270)
End

Procedure InitPiecePreviews()
Var
    i, size, x, y: Integer
    temp: Element
    tcol, tcol_a, wcol: Integer
Begin
    
    // Generate "previews"
    For i = 0 to (PP-1)
        size = IdToSize(i)
    
        temp = CreateImage(size*2, size*2)
        CLS(ToRGBA(0,0,255,255), temp)
        Rect(size/2, size/2, size, size, ToRGBA(255,255,255), TRUE, temp)
        previewPiece[i] = CreateSprite(size*2, size*2)
        SpriteVisible(previewPiece[i], FALSE)
        SetSpritePivot(previewPiece[i], size, size)
    
        ScaleTemplates(size, size)
        FlipTemplates()
        
        DrawTemplate(9, temp)
        DrawTemplate(10, temp)
        DrawTemplate(1, temp)
        DrawTemplate(2, temp)
        
        // Paint
        For x = 0 to size*2
            For y = 0 to size*2
                
                // Read Template Image
                tcol = Pixel(x, y, temp)
                tcol_a = Red(tcol)
                
                // Read Background Image
                
                    
                // Write Piece
                wcol = ToRGBA(255, 255, 255, tcol_a)
                Pixel(x, y, wcol, previewPiece[i])
                    
            Next
        Next
        
        FreeImage(temp)
        
        FreeFlippedTemplates()
        FreeScaledTemplates()
    Next  
End

Procedure FreePiecePreviews()
Var
    i: integer
Begin
    For i = 0 to (PP-1)
        FreeSprite(previewPiece[i])
    Next
End

Function IdToSize(i: Integer) : Integer
Var
    size: Integer
Begin

    Case i of
        0: size = 26
        1: size = 32
        2: size = 40
        3: size = 50
        4: size = 60
        5: size = 70
        6: size = 80
        7: size = 90
        8: size = 100
        9: size = 110
        10: size = 120
        11: size = 130
        12: size = 140
        
        Default:
            MessageBox("Invalid size id - defaulting to 80 pixels", "Warning")
            size = 80
    EndCase
    
    result = size
End

Function NewRotatedTemplateImage(img: Element, deg: Integer) : Element
Var
    w, h: Integer
    t: ELement
Begin
    deg = -deg
    w = ImageWidth(img)
    h = ImageHeight(img)
    result = CreateImage(w, h)
    t = CreateTransform()
    TransformRotate(t, w/2, h/2, deg)
    DrawTransform(img, t, 0, 0, result)
    FreeTransform(t)    
End

Procedure ScaleTemplates(w: Integer, h: Integer)
Var
    i: Integer
    t: Element
    pw, ph: Real
Begin
    t = CreateTransform()

    pw = ImageWidth(template[0])
    ph = ImageHeight(template[0])
    pw = ToReal(w)/pw
    ph = ToReal(h)/ph
    pw = pw / 2.0
    ph = ph / 2.0
    
    TransformScale(t, pw, ph)

    For i = 0 to (TI-1)
        template_scaled[i] = CreateImage(w/2, h/2)
        DrawTransform(template[i], t, 0, 0, template_scaled[i])
    Next
End

Procedure FreeScaledTemplates()
Var
    i: Integer
Begin
    For i = 0 to (TI-1)
        FreeImage(template_scaled[i])
    Next
End

Procedure FlipTemplates()
Var
    i: Integer
    pw, ph: Integer
    fx, fy: Boolean
Begin
    
    pw = ImageWidth(template_scaled[0])
    ph = ImageHeight(template_scaled[0])    

    For i = 0 to (TI-1)
    
        fx = FALSE
        fy = FALSE
        
        If ((i mod 2) = 0) then
            fx = TRUE
        Else
            fy = TRUE
        Endif
        
        template_flipped[i] = CreateImage(pw, ph)
        DrawImage(template_scaled[i], 0, 0, template_flipped[i])
        If (fx) then hFlipImage(template_flipped[i])
        If (fy) then vFlipImage(template_flipped[i])
        
    Next
End

Procedure FreeFlippedTemplates()
Var
    i: Integer
Begin
    For i = 0 to (TI-1)
        FreeImage(template_flipped[i])
    Next
End

Procedure DrawTemplate(id: Integer, target: Element)
Var
    xoffset, yoffset, xoffset2, yoffset2: Integer
    pw, ph, lid: Integer
Begin
    pw = ImageWidth(template_scaled[0])
    ph = ImageHeight(template_scaled[0])
    xoffset2 = 0
    yoffset2 = 0
    
    lid = id
    
    While (lid > 3)
        lid = lid - 4
    Wend
    
    Case id of
        4: yoffset2 = ph - 1
        5: xoffset2 = pw +1
        6: yoffset2 = ph
        7: xoffset2 = pw -2 ; yoffset2 = 1
        
        8: yoffset2 = ph*2-1
        9: xoffset2 = -pw*2 ; yoffset2 = 1
        10: yoffset2 = -ph*2
        11: xoffset2 = pw*2 ; yoffset2 = 1
    EndCase

    Case lid of
        0: xoffset = pw;     yoffset = 1
        1: xoffset = pw*3-1; yoffset = ph +1
        2: xoffset = pw*2-1; yoffset = ph*3 -2
        3: xoffset = 0;      yoffset = ph*2 -1
    EndCase

    DrawImage(template_scaled[id], xoffset+xoffset2, yoffset+yoffset2, target)
    
    Case lid of
        0: xoffset = pw*2;   yoffset = 1 
        1: xoffset = pw*3-1; yoffset = ph*2 -1
        2: xoffset = pw*1+1; yoffset = ph*3 -2
        3: xoffset = 0;      yoffset = ph +1
    EndCase
    
    DrawImage(template_flipped[id], xoffset+xoffset2, yoffset+yoffset2, target)
    
End

Procedure Main_JigSaw(puzzleFile: String, peiceSize: Integer)
Var
    done: Boolean = FALSE
    
    mx, my: Integer
    md, lastmd: Boolean
    ms, lastms: Integer

    puzzle, puzzleSprite: Element
    temp: Element
    t: Element
    
    x, y, i: Integer
    x2, y2: Integer
    xp, yp: Integer
       
    puzzlew, puzzleh, trimpuzzlew, trimpuzzleh : Integer
    xpieces, ypieces: Integer
    piecew, pieceh: Integer
    
    pa: Array [512, 512] of ^pieces
    activePiece: ^pieces
    existsActivePiece: Boolean
    
    tcol, wcol, pcol: Integer
    tcol_a, tcol_r, tcol_b: Integer
    
    pcol_r, pcol_g, pcol_b, pcol_a: Integer
    
    nx, ny, qx, qy: Integer
    solved: Boolean
    
    partsActive: Boolean
    
Begin
    // This procedure is quite monolithic, but it breaks into sections nicely
    
    puzzle = LoadImage("rsc/puzzles/"+puzzleFile)
    If (puzzle = NULL) then MessageBox("Error: Couldnt' load puzzle '"+puzzleFile+"'", "Error")
     
    puzzlew = ImageWidth(puzzle)
    puzzleh = ImageHeight(puzzle)
    
    xpieces = (puzzlew/peiceSize)-1
    ypieces = (puzzleh/peiceSize)-1
    piecew = peiceSize
    pieceh = peiceSize
    trimpuzzlew = xpieces * peiceSize
    trimpuzzleh = ypieces * peiceSize  
    
    ScaleTemplates(piecew, pieceh)
    FlipTemplates()
    
    // Plan Peices
    For x = 0 to (xpieces-1)
        For y = 0 to (ypieces-1)
        
            piece = NewItem(pieceList)
            pa[x,y] = piece
            
            piece.targetx = (x * piecew) - (piecew/2) + 24
            piece.targety = (y * pieceh) - (pieceh/2) + 24
            piece.posx = Rand(SW-284, SW-24-(piecew*2))
            piece.posy = Rand(24, SH-24-(pieceh*2))
            piece.ew = piecew*2
            piece.eh = pieceh*2
            piece.solved = FALSE 
            
            piece.e = CreateSprite(piecew*2, pieceh*2)
            PositionSprite(piece.e, piece.posx, piece.posy)
            
            piece.outie_r = (Rand(0, 1) = 1)
            piece.outie_d = (Rand(0, 1) = 1)

        Next
    Next
    
    // Finished Puzzle
    t = CreateTransform()
    TransformScale(t, ToReal(trimpuzzlew)/ToReal(puzzlew), ToReal(trimpuzzleh)/ToReal(puzzleh))
    puzzleSprite = CreateSprite(trimpuzzlew, trimpuzzleh)
    DrawTransform(puzzle, t, 0, 0, puzzleSprite)
    PositionSprite(puzzleSprite, 24, 24)
    SpriteVisible(puzzleSprite, FALSE)
    FreeTransform(t) 
    
    // Particles
    For i = 0 to (MAX_PARTS)
        part = NewItem(partList)
        part.e = CreateSharedSprite(particleMaster, TRUE)
        SpriteVisible(part.e, FALSE)
    Next

    // Cut + Paint Peices
    For x = 0 to (xpieces-1)
        For y = 0 to (ypieces-1)
        
            piece = pa[x,y]
            
            // Draw Template onto temporary image
            temp = CreateImage(piecew*2, pieceh*2)
            CLS(ToRGBA(0,0,255,255), temp)
            Rect(piecew/2, pieceh/2, piecew, pieceh, ToRGBA(255,255,255), TRUE, temp)
            
            // Handle straight-edge borders, outie and innie pieces. 
            If (y = 0) then
                DrawTemplate(4, temp)
            Else
                If (y > 0) and (not pa[x,y-1].outie_d) then
                    DrawTemplate(10, temp)
                Else
                    DrawTemplate(0, temp)
                Endif
            Endif

            If (x = 0) then
                DrawTemplate(7, temp)
            Else
                If (x > 0) and (not pa[x-1,y].outie_r) then
                    DrawTemplate(9, temp)
                Else
                    DrawTemplate(3, temp)
                Endif
            Endif
            
            If (y = ypieces-1) then
                DrawTemplate(6, temp)
            Else
                If (piece.outie_d) then
                    DrawTemplate(8, temp)
                Else
                    DrawTemplate(2, temp)
                Endif
            Endif
            
            If (x = xpieces-1) then
                DrawTemplate(5, temp)
            Else
                If (piece.outie_r) then
                    DrawTemplate(11, temp)
                Else
                    DrawTemplate(1, temp)
                Endif
            Endif
            
            // Paint
            For x2 = 0 to piecew*2
                For y2 = 0 to pieceh*2
                
                    // Read Template Image
                    tcol = Pixel(x2, y2, temp)
                    tcol_r = Red(tcol)
                    tcol_a = Red(tcol)
                    tcol_b = Blue(tcol)
                    
                    // Read Puzzle Image
                    xp = x2 + (x * piecew)
                    yp = y2 + (y * pieceh)
                    
                    If (xp < puzzlew) and (yp < puzzleh) then
                        pcol = Pixel(xp, yp, puzzle)
                        pcol_r = (ToReal(tcol_r)/255.0) * (ToReal(Red(pcol)))
                        pcol_g = (ToReal(tcol_r)/255.0) * (ToReal(Green(pcol)))
                        pcol_b = (ToReal(tcol_r)/255.0) * (ToReal(Blue(pcol)))
                    Else
                        pcol_r = 0 ; pcol_g = 0 ; pcol_b = 0
                    Endif
                    
                    // Write Piece
                    wcol = ToRGBA(pcol_r, pcol_g, pcol_b, tcol_a)
                    Pixel(x2, y2, wcol, piece.e)
                    
                Next
            Next
            
            FreeImage(temp)

            // Test
            //DrawImage(piece.e, piece.targetx, piece.targety, canvas)
            //Flip
       Next
    Next
    
    
    FreeFlippedTemplates()
    FreeScaledTemplates()
    
    // Canvas Background    
    CLS(ToRGBA(0,100,50), canvas)
    Rect(SW-284, 23, 260, SH-24-24+2, ToRGBA(0, 60, 0), TRUE, canvas)
    Rect(SW-284, 23, 260, SH-24-24+2, ToRGBA(200, 168, 50), FALSE, canvas)
    Rect(23, 23, trimpuzzlew+2, trimpuzzleh+2, ToRGBA(0, 60, 0), TRUE, canvas)
    
    // Loop Init
    md = FALSE
    existsActivePiece = FALSE
    solved = FALSE
    partsActive = FALSE
    ResetDelta()
    FlushKeys()
    lastms = Millisecs
    
    // Loop
    While (not done)
    
        lastmd = md
        lastms = ms
        ms = Millisecs
        md = MouseDown()
        mx = MouseX()
        my = MouseY()

        // Select a piece
        If (md) and (not lastmd) and (not existsActivePiece) then
            activePiece = SelectPiece(mx, my, existsActivePiece)
        Endif
        
        If existsActivePiece then
            nx = mx - (activePiece.ew/2)
            ny = my - (activePiece.eh/2)
            qx = (activePiece.ew/6)
            qy = (activePiece.eh/6)
        Endif
        
        // Move/drag selected piece
        If (md) and (existsActivePiece) then
            activePiece.posx = nx
            activePiece.posy = ny
            PositionSprite(activePiece.e, nx, ny)
        Endif
        
        // Drop a piece - attempt to fit
        If (not md) and (existsActivePiece) then
            existsActivePiece = FALSE
            
            If (nx >= activePiece.targetx - qx) and (nx <= activePiece.targetx + qx) then
                If (ny >= activePiece.targety - qy) and (ny <= activePiece.targety + qy) then
                
                    // Fit Piece
                    activePiece.posx = activePiece.targetx
                    activePiece.posy = activePiece.targety
                    PositionSprite(activePiece.e, activePiece.posx, activePiece.posy)
                    activePiece.solved = TRUE
                    
                    // Particles
                    If gParticles then
                        Explosion(activePiece.posx + (activePiece.ew/2), activePiece.posy + (activePiece.eh/2))
                        partsActive = TRUE
                    Endif
                    
                    // Check if the whole puzzle is solved
                    solved = IsSolved()
                    If solved then SpriteVisible(puzzleSprite, TRUE)
                    
                Endif
            Endif
        Endif
        
        // Wheeeee
        If partsActive then
            partsActive = UpdateExplosions(ms - lastms)
        Else
            If solved then
                Explosion(Rand(SW/6, SW - (SW/6)), Rand(SH/6, SH/2))
                partsActive = TRUE
            Endif
        Endif

        Flip
        Pause(1)
        
        If (KeyHits(VK_ESCAPE) > 0) then done = TRUE
        If (ExitRequested) then done = TRUE    
    Wend
    
    
    // Free
    Loop part through partList
        FreeSprite(part.e)
        Free(part)
    EndLoop
    
    Loop piece through pieceList
        FreeSprite(piece.e)
        Free(piece)
    EndLoop
    
    FreeImage(puzzle)
    FreeSprite(puzzleSprite)

    CLS(ToRGBA(0,0,0), canvas)
    Flip()    
    FlushKeys()

End

Function SelectPiece(x: Integer, y: Integer, ByRef existsActivePiece: Boolean): ^pieces
Begin    
    Loop piece back through pieceList
        If (x >= piece.posx + (piece.ew/4)) and (x <= piece.posx + (piece.ew - piece.ew/4)) then
            If (y >= piece.posy + (piece.eh/4)) and (y <= piece.posy + (piece.eh - piece.eh/4)) then
                If (not piece.solved) then
                    result = piece
                    existsActivePiece = TRUE
                    exit
                Endif
            Endif
        Endif
    EndLoop
End

Function IsSolved() : Boolean
Begin
    result = FALSE

    Loop piece through pieceList
        If (not piece.solved) then exit
    EndLoop
    
    result = TRUE
End

Procedure Explosion(x: Integer, y: Integer)
Begin
    Loop part through partList
        part.xpos = x
        part.ypos = y
        part.xvol = 0
        part.yvol = 0
        part.ealpha = 1.0                
        part.size = Rand(8, 64)
        part.rot = Rnd(0, 1.0)
        part.active = TRUE
        
        While (Abs(part.xvol)<0.4)
            part.xvol = Rnd(-4.0, 4.0)
        Wend
        
        While (Abs(part.yvol)<0.4)
            part.yvol = Rnd(-4.0, 4.0)
        Wend
        
        SpriteVisible(part.e, TRUE)
        SpriteAlpha(part.e, 255)
        SpriteAngle(part.e, part.rot*360)
        SpriteWidth(part.e, part.size)
        SpriteHeight(part.e, part.size)
        SetSpritePivot(part.e, part.size/2, part.size/2)
        PositionSprite(part.e, part.xpos, part.ypos) 
    EndLoop
End

Function UpdateExplosions(diff: Integer) : Boolean
Var
    delta: Real
    c: Integer
Begin
    delta = ToReal(diff)/1000.0
    delta = NormaliseDelta(delta)
    c = 0
    result = TRUE

    Loop part through partList
        If (part.active) then
            Inc(c)
        
            part.rot = part.rot + delta
            part.xpos = part.xpos + (delta * part.xvol * 200)
            part.ypos = part.ypos + (delta * part.yvol * 200)
            part.yvol = part.yvol + (delta * 8)
            part.ealpha = part.ealpha - delta
             
            If (part.ealpha < 0) then
                part.ealpha = 0
                part.active = FALSE
            Endif
            
            If (part.xpos < 0 - (part.size/2)) then part.active = FALSE
            If (part.xpos > SW + (part.size/2)) then part.active = FALSE
            If (part.ypos > SH + (part.size/2)) then part.active = FALSE

            SpriteAngle(part.e, part.rot*360)
            SpriteAlpha(part.e, part.ealpha*255)
            PositionSprite(part.e, part.xpos, part.ypos)
            
        Else
            SpriteVisible(part.e, FALSE)
        Endif        
    EndLoop
    
    If (c = 0) then result = FALSE
End

Procedure Splash(logoFile: String, delay: Integer, inOnly: Boolean=FALSE)
Var
    logo, slogo, t: Element
    s, a: Real
    w, h, x: Integer
    done: Boolean = FALSE
    start, m, diff, dir: Integer
Begin

    // Yay!
    logo = LoadImage("rsc/"+logoFile)
    If (logo = NULL) then
        MessageBox("Couldn't load splash '"+logoFile+"'", "Error")
        exit
    Endif 
    
    s = ToReal(SH) / ToReal(ImageHeight(logo))
    w = ImageWidth(logo) * s
    h = ImageHeight(logo) * s 
    x = (SW-w)/2
    
    slogo = CreateSprite(w, h)
    t = CreateTransform()
    
    TransformScale(t, s, s)
    DrawTransform(logo, t, x, 0, slogo)
    SpriteAlpha(logo, 0)
    a = 0
    dir = 1
    start = Millisecs

    While (not done)
    
        m = Millisecs
        diff = (m - start)

        a = ToReal(diff) / 1000.0 * (255.0)
        If (a > 255) then
            a = 255
            
            If dir = 1 then
                dir = -1
                a = 0
                PauseGood(delay)
                start = Millisecs
                if (inOnly) then break
            Else
                done = TRUE 
            Endif
        Endif
        
        If dir = 1 then
            SpriteAlpha(slogo, a)
        Else
            SpriteAlpha(slogo, 255-a)
        Endif 
    
        Flip
        Pause(1)
    Wend
    
    FreeImage(logo)
    FreeSprite(slogo)
    
End

Procedure PauseGood(m: Integer)
Var
    start: Integer
Begin
    start = Millisecs
    
    While TRUE
        If ((Millisecs - start) >= m) then exit
        Flip
        Pause(1)
    Wend
End

Procedure ResetDelta()
Begin
    NormaliseDelta(0.0, TRUE)
End

Function NormaliseDelta(fd: Real, reset: Boolean = FALSE) : Real
Var
    avg, fd1, fd2, fd3, fd4, fd5, fd6, fd7, fd8, fd9: Static Real
Begin

    If reset then
        fd1 = 0.0
        fd2 = 0.0
        fd3 = 0.0
        fd4 = 0.0
        fd5 = 0.0
        fd6 = 0.0
        fd7 = 0.0
        fd8 = 0.0
        fd9 = 0.0
        exit
    Endif

    avg = (fd1 + fd2 + fd3 + fd4 + fd5 + fd6 + fd7 + fd8 + fd9 + fd) / 10.0 
    
    fd1 = fd2
    fd2 = fd3
    fd3 = fd4
    fd4 = fd5
    fd5 = fd6
    fd6 = fd7
    fd7 = fd8
    fd8 = fd9
    fd9 = fd
    
    If fd > avg then fd = avg
    result = fd

End

Procedure DarkenSprite(src: Element, dest: Element, d: Integer)
Var
    x, y, w, h: Integer
    col, r, g, b: Integer
Begin
    w = SpriteWidth(dest)
    h = SpriteHeight(dest)
    
    For y = 0 to h-1
        For x = 0 to w-1
        
            col = Pixel(x, y, src)
            
            r = Red(col)
            g = Green(col)
            b = Blue(col)
            
            r = ToReal(r) * (ToReal(d)/255.0)
            g = ToReal(g) * (ToReal(d)/255.0)
            b = ToReal(b) * (ToReal(d)/255.0)
            
            Pixel(x, y, ToRGBA(r,g,b), dest)
            
        Next
    Next
End

Procedure BlingifyPuzzleBox(src: Element, dest: Element, d: Integer=255)
Var
    x, y: Integer
    w, h, w2, h2: Integer
    r, g, b: Integer
    color: Integer
    a: Integer
Begin

    w = ImageWidth(corner)
    h = ImageHeight(corner)
    w2 = ImageWidth(src)
    h2 = ImageHeight(src)

    // top-left
    For y = 0 to (h-1)
        For x = 0 to (w-1)
            a = Red(Pixel(x, y, corner))
            color = Pixel(x, y, src)
            r = ToReal(Red(color)) * (ToReal(d)/255.0) 
            g = ToReal(Green(color)) * (ToReal(d)/255.0)
            b = ToReal(Blue(color)) * (ToReal(d)/255.0)
            Pixel(x, y, ToRGBA(r, g, b, a), dest)
        Next
    Next
    
    // top-mid
    For y = 0 to (h-1)
        For x = w to (w2-w)
            color = Pixel(x, y, src)
            r = ToReal(Red(color)) * (ToReal(d)/255.0) 
            g = ToReal(Green(color)) * (ToReal(d)/255.0)
            b = ToReal(Blue(color)) * (ToReal(d)/255.0)
            Pixel(x, y, ToRGBA(r, g, b), dest)
        Next
    Next
    
    // top-right
    For y = 0 to (h-1)
        For x = 0 to (w-1)
            a = Red(Pixel(x, y, corner))
            color = Pixel(w2 - x, y, src)
            r = ToReal(Red(color)) * (ToReal(d)/255.0) 
            g = ToReal(Green(color)) * (ToReal(d)/255.0)
            b = ToReal(Blue(color)) * (ToReal(d)/255.0)
            Pixel(w2 - x, y, ToRGBA(r, g, b, a), dest)
        Next
    Next 
    
    // mid
    For y = h to (h2-1)
        For x = 0 to (w2-1)
            color = Pixel(x, y, src)
            r = ToReal(Red(color)) * (ToReal(d)/255.0) 
            g = ToReal(Green(color)) * (ToReal(d)/255.0)
            b = ToReal(Blue(color)) * (ToReal(d)/255.0)
            Pixel(x, y, ToRGBA(r, g, b), dest)
        Next
    Next 

End

Procedure DrawPuzzleReflection(src: Element, dest: Element)
Var
    x, y, w, h, hsrc: Integer
    col, r, g, b, a, d: Integer
    xoffset: Integer
Begin
    w = ImageWidth(src)
    hsrc = ImageHeight(src)
    h = SH/3
    d = 100
    
    For y = 0 to h-1 step 4
        For x = 0 to w-1
        
            col = Pixel(x, (hsrc-y), src)
            
            r = Red(col)
            g = Green(col)
            b = Blue(col)
            
            r = ToReal(r) * (ToReal(d)/255.0)
            g = ToReal(g) * (ToReal(d)/255.0)
            b = ToReal(b) * (ToReal(d)/255.0)
            a = (1.0 - ToReal(y)/ToReal(h)) * 200.0
            
            xoffset = (1.0 - ToReal(y)/ToReal(h)) * (w/2) 
            
            Pixel(x+xoffset, y/4, ToRGBA(r,g,b,a), dest)
            
        Next
    Next
End

Function Main_SelectPuzzle() : String
Var
    done: Boolean = FALSE
    
    mx, my: Integer
    md, lastmd: Boolean
    ms, lastms: Integer
    delta: Real
    
    dir: Element
    temp, temp2, temp3, temp4: Element
    t: Element
    filename: String
    
    xoffset, xaccel: Real
    x, y: Integer
    
    listwidth: Real
    c, cnt: Integer
    resizeScalar, resizeScalarW: Real
    
    firstx, lastx: Integer
    
    activePuzzle: ^puzzles
    existsActivePuzzle: Boolean
    
    table, escToExit: Element
Begin
    result = ""
    CLS(ToRGBA(0,0,0), canvas)
    
    // Esc to Exit
    escToExit = LoadSprite("rsc/esc_to_exit.png")
    PositionSprite(escToExit, SW-10-ImageWidth(escToExit), 10)
    
    // Table
    temp = LoadImage("rsc/table.png")
    resizeScalar = ToReal(SH/3) / ToReal(ImageHeight(temp))
    resizeScalarW = ToReal(SW) / ToReal(ImageHeight(temp))
    t = CreateTransform()
    TransformScale(t, resizeScalarW, resizeScalar)
    table = CreateSprite(SW, SH/3)
    PositionSprite(table, 0, SH-SH/3)
    DrawTransform(temp, t, 0, 0, table)
    FreeTransform(t)
    FreeImage(temp)
    
    
    // Scan directory for puzzles
    dir = OpenDir("rsc/puzzles/*.jpg", CFA_READONLY + CFA_HIDDEN)
    filename = NextFile(dir)
    While (filename <> "")
        
        puzzle = NewItem(puzzleList)
        
        temp = LoadImage("rsc/puzzles/"+filename)
        If (temp = NULL) then MessageBox("Error loading '"+filename+"'", "Error"); exit
        t = CreateTransform()
        
        resizeScalar = ToReal(SH - SH/3) / ToReal(ImageHeight(temp))
        TransformScale(t, resizeScalar, resizeScalar)
        
        puzzle.filename = filename
        puzzle.w = ImageWidth(temp) * resizeScalar
        puzzle.h = ImageHeight(temp) * resizeScalar
               
        puzzle.e            = CreateSprite(puzzle.w, puzzle.h)
        puzzle.e_hover      = CreateSprite(puzzle.w, puzzle.h)
        puzzle.e_reflection = CreateSprite(puzzle.w*1.5, SH/3) 
        
        temp2 = CreateImage(puzzle.w, puzzle.h)
        temp3 = CreateImage(puzzle.w, puzzle.h)
        temp4 = CreateImage(puzzle.w, puzzle.h)
        
        puzzle.ealpha = 0.0
        
        If (puzzle.e = NULL) then MessageBox("Error: puzzle.e is NULL", "Error")
        If (puzzle.e_hover = NULL) then MessageBox("Error: puzzle.e_hover is NULL", "Error")
        
        DrawTransform(temp, t, 0, 0, temp2)
        
        BlingifyPuzzleBox(temp2, temp3)
        BlingifyPuzzleBox(temp2, temp4, 128)
        
        DrawImage(temp4, 0, 0, puzzle.e)
        DrawImage(temp3, 0, 0, puzzle.e_hover)
        DrawPuzzleReflection(temp3, puzzle.e_reflection)
        
        SetSpritePivot(puzzle.e, puzzle.w/2, puzzle.h)
        SetSpritePivot(puzzle.e_hover, puzzle.w/2, puzzle.h)
        SetSpritePivot(puzzle.e_reflection, puzzle.w, 0)
        SpriteAlpha(puzzle.e_hover, 0)
        
        FreeTransform(t)
        FreeImage(temp4)
        FreeImage(temp3)
        FreeImage(temp2)
        FreeImage(temp)  
    
        filename = NextFile(dir)
    Wend
    
    // Calculate width of list
    listwidth = 0
    c = 0
    cnt = CountList(puzzleList)
    Loop puzzle through puzzleList
        puzzle.x = listwidth + puzzle.w/2 
        listwidth = listwidth + puzzle.w + PZL_PADDING
        
        If c = 0 then firstx = puzzle.w + PZL_PADDING - puzzle.w/2
        If c = cnt-1 then lastx = listwidth - SW - PZL_PADDING + puzzle.w/4
        
        Inc(c)
    EndLoop
    listwidth = listwidth - PZL_PADDING
    
    If (c < 2) then
        MessageBox("Not enough puzzles!!!")
        exit
    Endif 
 
    
    // Loop Init
    xoffset = -(SW/2)
    FlushKeys()
    SetMousePos(SW/2, SH/2)
    
    // Loop
    md = FALSE
    done = FALSE
    xaccel = 0
    existsActivePuzzle = FALSE
    ResetDelta()
    FlushKeys()
    lastms = Millisecs
    
    // Loop
    While (not done)
    
        lastmd = md
        lastms = ms
        ms = Millisecs
        md = MouseDown()
        mx = MouseX()
        my = MouseY()
        delta = ToReal(ms - lastms)/1000.0
        delta = NormaliseDelta(delta)
        
        // Move Left/Right +accelerate
        Conditions
            (mx > (SW - SW/3)):
                If xaccel < 1.0 then xaccel = 100.0
                xoffset = xoffset + (delta * xaccel)
                xaccel = xaccel + (delta * 1000)
                If xaccel > 750 then xaccel = 750
            (mx < (SW/3)):
                If xaccel < 1.0 then xaccel = 100.0
                xoffset = xoffset - (delta * xaccel)
                xaccel = xaccel + (delta * 1000)
                If xaccel > 750 then xaccel = 750
            default:
                xaccel = 0
        EndConditions
        
        // Limit
        If xoffset < -firstx then xoffset = -firstx
        If xoffset > lastx then xoffset = lastx
        
        // Update Puzzles for display
        Loop puzzle through puzzleList
            x = puzzle.x - xoffset
            y =  SH - (SH/4)
            
            // Mouse Aware
            If (mx >= x-puzzle.w/2) and (mx < x+puzzle.w/2) and (my >= y-puzzle.h) and (my <= y) then
                existsActivePuzzle = TRUE
                activePuzzle = puzzle
            Else
                If (existsActivePuzzle) and (activePuzzle = puzzle) then
                    existsActivePuzzle = FALSE
                Endif
            Endif
            
            If (existsActivePuzzle) and (activePuzzle = puzzle) then
                puzzle.ealpha = puzzle.ealpha + (delta * 2) 
            Else
                puzzle.ealpha = puzzle.ealpha - (delta * 2)
            Endif
            If puzzle.ealpha < 0 then puzzle.ealpha = 0
            If puzzle.ealpha > 1 then puzzle.ealpha = 1
            
            If (md) and (existsActivePuzzle) and (activePuzzle = puzzle) then
                done = TRUE
                result = activePuzzle.filename
            Endif 
        
            // Position
            PositionSprite(puzzle.e, x, y)
            PositionSprite(puzzle.e_hover, x, y)
            PositionSprite(puzzle.e_reflection, x, y)
            SpriteAlpha(puzzle.e_hover, puzzle.ealpha*255.0)
            
            
        EndLoop
        
        Flip
        Pause(1)
        
        If (KeyHits(VK_ESCAPE) > 0) then done = TRUE
        If (ExitRequested) then done = TRUE 
    Wend
    
    // Free
    Loop puzzle through puzzleList
        FreeSprite(puzzle.e)
        FreeSprite(puzzle.e_hover)
        FreeSprite(puzzle.e_reflection)
        Free(puzzle)
    EndLoop
    
    FreeSprite(table)
    FreeSprite(EscToExit)
    
    FlushKeys()
    
End

Procedure InitButtons()
Var
    i, j: Integer
Begin

    // Buttons
    btn[0,0] = LoadSprite("rsc/btn/smaller.png")
    btn[0,1] = LoadSprite("rsc/btn/smaller_hover.png")
    btn[0,2] = LoadSprite("rsc/btn/smaller_active.png")
    btn[1,0] = LoadSprite("rsc/btn/bigger.png")
    btn[1,1] = LoadSprite("rsc/btn/bigger_hover.png")
    btn[1,2] = LoadSprite("rsc/btn/bigger_active.png")
    btn[2,0] = LoadSprite("rsc/btn/start.png")
    btn[2,1] = LoadSprite("rsc/btn/start_hover.png")
    btn[2,2] = LoadSprite("rsc/btn/start_active.png")

    For i = 0 to 2
        For j = 0 to 2
            SetSpritePivot(btn[i,j], SpriteWidth(btn[i,j])/2, SpriteHeight(btn[i,j])/2)
        Next
    Next    
    
    HideButtons()
    
End

Procedure HideButtons()
Var
    i, j: Integer
Begin
    For i = 0 to 2
        For j = 0 to 2
            SpriteVisible(btn[i, j], FALSE)
        Next
    Next
End

Procedure FreeButtons()
Var
    i, j: Integer
Begin
    For i = 0 to 2
        For j = 0 to 2
            FreeSprite(btn[i, j])
        Next
    Next
End

Function Main_SelectSize() : Integer
Var
    done: Boolean = FALSE
    i: Integer
    j: Integer
    x, y, w, h: Integer
    
    mx, my: Integer
    mh, md, lastmd: Boolean
    
    currentSize: Integer = 6
    
    btnx: Array[3] of Integer
    btny: Array[3] of Integer
    vis: Integer
Begin

    CLS(ToRGBA(0,0,0), canvas)
    FlushKeys()
    
    // Peices
    x = 0; y = 0;
    For i = 0 to (PP-1)
        SpriteVisible(previewPiece[i], FALSE)
        PositionSprite(previewPiece[i], SW/2, SH/2) 
    Next
    SpriteVisible(previewPiece[currentSize], TRUE)
    
    // Buttons
    btnx[0] =  SW/2 - 128
    btny[0] =  SH/2
    btnx[1] =  SW/2 + 128
    btny[1] =  SH/2
    btnx[2] =  SW/2
    btny[2] =  SH/2+128
    w = SpriteWidth(btn[0,0])
    h = SpriteHeight(btn[0,0]) 
    
    For i = 0 to 2
        For j = 0 to 2            
            PositionSprite(btn[j,i], btnx[j], btny[j])
        Next
    Next
    
    // Loop
    md = FALSE
    While (not done)
        
        lastmd = md
        mx = MouseX()
        my = MouseY()
        md = MouseDown()
        mh = (not md) and (lastmd)
        
        // Update Buttons
        HideButtons()
        
        For i = 0 to 2
            
            If (mx > btnx[i] - w/2) and (mx < btnx[i] + w/2) and (my > btny[i] - h/2) and (my < btny[i] + h/2) then
                vis = 1
                If md then vis = 2
                
                // Action
                If mh then
                    Case i of
                        0:
                            Dec(currentSize)
                            If (currentSize < 0) then currentSize = 0
                        1:
                            Inc(currentSize)
                            If (currentSize = PP) then currentSize = PP-1
                        2:
                            done = TRUE
                    EndCase
                    
                    For j = 0 to (PP-1)
                        SpriteVisible(previewPiece[j], FALSE)
                    Next
                    SpriteVisible(previewPiece[currentSize], TRUE)
                Endif
            Else
                vis = 0
            Endif
            
            SpriteVisible(btn[i,vis], TRUE)
        
        Next 
    
    
    
        Flip
        Pause(1)
        
        If (KeyHits(VK_ESCAPE) > 0) then done = TRUE
    Wend
    
    // Pieces
    For i = 0 to (PP-1)
        SpriteVisible(previewPiece[i], FALSE)
    Next
    
    HideButtons()
    result = IdToSize(currentSize)
        
End

Procedure Main()
Var
    done: Boolean = FALSE
    puzzle: String
    size: Integer
Begin
    
    While (done = FALSE)
        
        Splash("loading-puzzle.png", 0, TRUE)
        puzzle = Main_SelectPuzzle()
        If (puzzle = "") then exit
        
        Splash("select_size.png", 1000, TRUE)
        size = Main_SelectSize()
    
        Splash("cutting.png", 0, TRUE)
        Main_JigSaw(puzzle, size)

        Flip
        Pause(1)
        
        If (ExitRequested) then done = TRUE  
    Wend
End

Begin
    SetAppName("Christmas Jigsaw")
    Randomize(Millisecs)
    InitCfg()
    InitScreen()
    InitTemplate()
    InitPiecePreviews()
    InitButtons()
    
    If gShowSplash then Splash("splash_jigsaw.png", 3000)
    If gShowSplash then Splash("xmasxompo_1024.jpg", 2250)
    
    particleMaster = LoadSprite("rsc/particle.png") ; SpriteVisible(particleMaster, FALSE)
    corner = LoadImage("rsc/corner.png")
    
    Main()
    
    FreeButtons()
    FreePiecePreviews()
    FreeImage(corner)
    FreeSprite(particleMaster)
    FreeSprite(canvas)
End